/* A replacement for the original version.  This does not avoid
   comments and strings, and uses different algorithms.  The
   hashlib package is a prerequisite, and supplies the storage.

   This is a complete rewrite, in a different language and thus
   begins life as version 2.0 of id2id.

   Id2id reads a list of idpairs in the form of lines holding
       oldname newname
   and then applies all those changes to the file specified. It
   does this with a single pass through the file, and thus is
   almost as quick as simply copying the file.

   Unlike a normal text replacement in an editor, id2id only
   operates on complete identifiers.  Thus it cannot be used
   to replace prefixes or suffixes.  An identifier is anything
   that begins with '_' or any letter, and continues for all
   following characters until a non '_', letter, or digit.

   The primary purpose is to make compatible changes across a
   complete set of files.  The process of controlling the files
   modified is left for a supervisory script.

   Another use is when learning another program.  Identifiers
   can be changed to something more understandable.  As more
   and more of the program is understood, the idpairs file can
   be enhanced, always maintaining the original source unchanged

   This can also be applied to the output of a disassembler.

   The -r switch reverses the meanings of the columns of the
   idpairs file, and thus can often be used to recreate the
   original file.  BEWARE - this is not always possible.  In
   particular it is not possible if a 'newname' already existed
   in the original file.

   Copyright (C) 2004 by C.B. Falconer

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public
   License along with this program; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA  02111-1307, USA

   I can be contacted by mailto:cbfalconer@worldnet.att.net
           or at 680 Hartford Tpk., Hamden, CT 06517, USA.

   At the time of writing, this package was available at:

     <http://cbfalconer.home.att.net/download/id2id-20.zip>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iso646.h>
#include "hashlib.h"

/* thisword, otherword, from the point of view of the table */
/* i.e. the pointers will be reversed in the other table    */
typedef struct id {
   struct id *next;            /* utility, for forming lists */
   char      *thisword;        /* identifier to replace */
   char      *otherword;       /* what to replace it with */
   int        seeninsource;    /* newword found in source */
   int        seenasold;       /* newword seen as oldword */
   int        entrycount;      /* incase multiple entries */
} id, *idptr;

hshtbl *oldtable;              /* keys on oldwords */
hshtbl *newtable;              /* keys on newwords */

/* IMPORTANT - the strings are stored simultaneously in
   both oldtable and newtable, but only one is considered
   to belong to each.  Thus both tables must be open for
   any entries to be made to either.
*/

char idpairs[FILENAME_MAX] = "idpairs"; /* default */
int  reverse;                           /* default 0 */

FILE *idp;                     /* stream for idpairs loading */

/* controls maximum identifier length handled */
#define BUFMAX 40

/* ----------------- */

/* This is very likely to be non-portable
   DOES NOT check fp open for reading
   NULL fp is considered a keyboard here!
   With "gcc -W -Wall -ansi -pedantic" we can expect
   implicit declaration warnings for isatty and fileno.
   However the result should link correctly.
   If it always returns 0 the system still works
   but will not give the automatic help.
*/
static int akeyboard(FILE *fp)
{
#ifndef __TURBOC__  /* TC2 is peculiar */
# ifdef __STDC__
  /* This dirty operation allows gcc -ansi -pedantic */
  extern int fileno(FILE *fp);
  extern int isatty(int fn);
# endif
#endif
   return ((fp != NULL) && isatty(fileno(fp)));
} /* akeyboard */

/* ----------------- */

/* usage and error message(s) and exit */
static void help(const char * msg, const char *msg2)
{
   if (msg)  fputs(msg,  stderr);
   if (msg2) fputs(msg2, stderr);
   fputs("\n"
         "Usage: id2id [options] [idpairs] <infile >outfile\n"
         " infile and outfile default to stdin and stdout\n"
         " and idpairs defaults to \"idpairs\"\n"
         "\n"
         "Options include:\n"
         " -h                for this help\n"
         " -r                to reverse idpairs order\n"
         , stderr);
   exit(EXIT_FAILURE);
} /* help */

/* ----------------- */

/* Scan for options anywhere in command line */
/* Options always begin with '-' */
/* Arguments always DO NOT begin with '-' */
static int getopts(int argc, char **argv)
{
   int   i, n;
   char *p;

   for (i = 1, n = 0; i < argc; i++) {
      if ('-' == *(argv[i])) {
         p = 1 + argv[i];
         switch (tolower((unsigned char) *p++)) {
case 'r':   reverse = 1; break;
case 'h':   help(NULL, NULL); break;
default:    help("unknown option\n", NULL); break;
         } /* switch */
      }
      else {
         n++; /* advance to next argument */
         if (1 == n) strcpy(idpairs, argv[i]);
         else help("Too many arguments\n", NULL);
      }
   }
   return 1;
} /* getopts */

/* ----------------- */

/* skips and echoes all characters that cannot start an
   identifier. echoto may be NULL, to suppress any echoing
   Returns the starting char for the following identifier,
   which may be EOF and is still in the input stream.
*/
static int skiptoident(FILE *f, FILE *echoto)
{
   int   ch;

   do {
      ch = getc(f);
      if (('_' == ch) || isalpha(ch)) break;
      if (echoto && (EOF != ch)) putc(ch, echoto);
   } while (EOF != ch);
   ungetc(ch, f);
   return ch;
} /* skiptoident */

/* ----------------- */

/* generic function to extract the next identifier from a stream
   Identifiers can begin with _ or alpha, and continue until a
   non-alpha non _ non-numeric char is encountered.
   '\0' terminate the string.
   Leading chars that do not fit an identifier are echoed.
   echoto may be NULL, to suppress any echoing
   Returns char terminating the identifier, which is 'ungotten'
   buf[bmax] should be valid storage space.
*/
static int getident(char *buf, int bmax, FILE *f, FILE *echoto)
{
   int     ch, ix;

   if (EOF == skiptoident(f, echoto)) ch = EOF;
   else ch = getc(f);
   /* Here the input stream has been absorbed and echoed */
   /* up to the first character of an identifier, in ch  */
   ix = 0;
   while (('_' == ch) || isalnum(ch)) {  /* skips on eof */
      buf[ix++] = ch;
      ch = getc(f);
      if (ix >= bmax) break;
   }
   buf[ix] = '\0';
   ungetc(ch, f);
   return ch;
} /* getident */

/* ----------------- */

/* since '\n' is a terminator, and our input routines always  */
/* leave terminators in the input stream, this can flush them */
/* In the process we allow comments after the second word in  */
/* an idpairs line. */
static int flushln(FILE *f)
{
   int ch;

   while (('\n' != (ch = getc(f))) && (EOF != ch)) continue;
   return ch;
} /* flushln */

/* ----------------- */

/* These routines are the service routines for the two
   databases.  The difference will be in the data inserted.
*/

/* this is of type hshfn described in hashlib.h */
static unsigned long idhashfn(void *p)
{
   idptr idp = p;

   return hshstrhash(idp->thisword);
} /* idhashfn */

/* ----------------- */

/* this is of type hshfn described in hashlib.h */
static unsigned long idrehashfn(void *p)
{
   idptr idp = p;

   return hshstrehash(idp->thisword);
} /* idrehashfn */

/* ----------------- */

/* this is of type hshcmpfn described in hashlib.h */
static int idcompare(void *left, void *right)
{
   idptr lidp = left, ridp = right;

   return strcmp(lidp->thisword, ridp->thisword);
} /* idcompare */

/* ----------------- */

/* this is of type hshdupfn described in hashlib.h */
static void *iddupe(void *p)
{
   idptr idp = p;
   idptr dupid;

   if ((dupid = malloc(sizeof *dupid))) {
      if ((dupid->thisword = malloc(1 + strlen(idp->thisword)))) {
         strcpy(dupid->thisword, idp->thisword);
         dupid->otherword    = NULL;
         dupid->next      = NULL;
         dupid->seeninsource = 0;
         dupid->seenasold = 0;
         dupid->entrycount = 0;
      }
      else {
         free(dupid);
         dupid = NULL;
      }
   }
   return dupid;
} /* iddupe */

/* ----------------- */

/* this is of type hshfreefn described in hashlib.h */
static void idundupe(void *p)
{
   idptr idp = p;

   free(idp->thisword);
   free(p);
} /* idundupe */

/* End of functions that customize the hashtable system */

/* ----------------- */

/* This prepares two tables, one holding words to be replaced
   in the source file, and the other holding replacement words
   for those.  The actual strings are stored only once.  This
   symmetry enables the easy inversion of the changes (but
   beware if any new words appear in the old word table). The
   point of the hashtable is that any identifier can be found
   very quickly, if present.  The search is O(1).
*/
static void readidpairs(void)
{
   char   leftwd[BUFMAX+1], rightwd[BUFMAX+1];
   id     newentry, oldentry;
   int    ch;
/* linecount = 0; */              /* just for testing */
   id    *newwordp, *oldwordp;  /* table storage locations */
   hshtbl *temp;

   /* The tables are distinct. Different things go in them */
   oldtable = hshinit(idhashfn, idrehashfn,
                      idcompare,
                      iddupe, idundupe,
                      0);
   newtable = hshinit(idhashfn, idrehashfn,
                      idcompare,
                      iddupe, idundupe,
                      0);

   if (!oldtable || !newtable) {
      fprintf(stderr, "\n%s\n",
              "Can't secure memory for tables");
      exit(EXIT_FAILURE);
   }

   oldentry.thisword = leftwd;
   newentry.thisword = rightwd;
   do {
      ch = getident(leftwd, BUFMAX, idp, NULL);
      ch = getident(rightwd, BUFMAX, idp, NULL);
      ch = flushln(idp);
      oldwordp = hshinsert(oldtable, &oldentry);
      newwordp = hshinsert(newtable, &newentry);
      if (oldwordp && newwordp) {
         /* succesful insertion in both tables */
         oldwordp->entrycount++;
         oldwordp->otherword = newwordp->thisword;
         newwordp->entrycount++;
         newwordp->otherword = oldwordp->thisword;
         /* Later - warnings if entrycount > 1 */
      }
      else {  /* failure */
         fprintf(stderr, "\n,%s\n",
                 "Failure during name insertion");
         exit(EXIT_FAILURE);
      }
/*    fprintf(stderr, "%s %s\n", leftwd, rightwd);
      linecount++; */
   } while (EOF != ch);
/* fprintf(stderr, "%d lines read\n", linecount); */

   if (reverse) { /* invert table usage */
      temp = oldtable;
      oldtable = newtable;
      newtable = temp;
   }
} /* readidpairs */

/* ----------------- */

/* Copy stdin to stdout, replacing all identifiers in the
   oldtable list with the corresponding one in the newtable
   list.  This is only a single pass through the file.  This
   comment is written as if oldtable and newtable were the
   original columns in idpairs file.
*/
static void filter(void)
{
   char buf[BUFMAX+1];
   id   entry, *change, *check;

   entry.thisword = buf;

   while (EOF != skiptoident(stdin, stdout)) {
      getident(buf, BUFMAX, stdin, NULL);
      if (!(change = hshfind(oldtable, &entry)))
         printf("%s", buf);
      else {
         printf("%s", change->otherword);
         change->seeninsource = 1;
         if ((check = hshfind(newtable, &entry))) {
            check->seeninsource = 1;
         }
      }
   }
} /* filter */

/* ----------------- */

int main(int argc, char **argv)
{
   if (akeyboard(stdin)) help("Keyboard found", NULL); /* exit */
   if (not getopts(argc, argv)) help(NULL, NULL); /* and exit */
   if (not(idp = fopen(idpairs, "r")))
      help("can't open ", idpairs);
   oldtable = newtable = NULL;
   readidpairs();
   fclose(idp);
   filter();
   return 0;
} /* main, id2id */
